@@ -1,6 +1,6 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
  *
- * Copyright (C) 2007 Tadas Dailyda <tadas@dailyda.com>
+ * Copyright (C) 2007-2008 Tadas Dailyda <tadas@dailyda.com>
  *
  * Licensed under the GNU General Public License Version 2
  *
@@ -166,8 +166,10 @@ ods_bluez_finalize (GObject *object)
 
 	g_return_if_fail (bluez->priv != NULL);
 	
-	g_object_unref (G_OBJECT (bluez->priv->manager_proxy));
-	g_object_unref (G_OBJECT (bluez->priv->adapter_proxy));
+	if (G_IS_OBJECT (bluez->priv->manager_proxy))
+		g_object_unref (G_OBJECT (bluez->priv->manager_proxy));
+	if (G_IS_OBJECT (bluez->priv->adapter_proxy))
+		g_object_unref (G_OBJECT (bluez->priv->adapter_proxy));
 
 	G_OBJECT_CLASS (ods_bluez_parent_class)->finalize (object);
 }
@@ -350,24 +352,38 @@ get_remote_service_handles_cb (DBusGProxy *proxy, DBusGProxyCall *call,
 	gboolean		ret;
 	GError			*error = NULL;
 	GArray			*handle_array = NULL;
-	guint32			service_handle;
+	guint32			service_handle = 0;
 	
 	ret = dbus_g_proxy_end_call (proxy, call, &error,
 								DBUS_TYPE_G_UINT_ARRAY, &handle_array,
 								G_TYPE_INVALID);
 	
+	/* check if we were looking for Nokia specific FTP service and failed */
+	if (ret && handle_array->len == 0 && !strcmp (cb_data->uuid, OBEX_NOKIAFTP_UUID)) {
+		g_free (cb_data->uuid);
+		cb_data->uuid = g_strdup (OBEX_FTP_UUID);
+		dbus_g_proxy_begin_call (proxy, 
+								"GetRemoteServiceHandles",
+								(DBusGProxyCallNotify) get_remote_service_handles_cb,
+								cb_data, NULL,
+								G_TYPE_STRING, cb_data->address,
+								G_TYPE_STRING, cb_data->uuid,
+								G_TYPE_INVALID);
+		return;
+	}
+	/* service search failed */
 	if (!ret || handle_array->len == 0) {
-		/* It means that Host is down */
 		g_clear_error (&error);
 		g_set_error (&error, ODS_ERROR, ODS_ERROR_CONNECTION_ATTEMPT_FAILED,
-						"Host down");
+						"Service search failed");
 		cb (-1, error, cb_data->data);
 		cb_data_free (cb_data);
 		g_clear_error (&error);
 		goto out;
 	}
-	service_handle = *((guint32 *)handle_array->data);
-	
+
+	memcpy(&service_handle, handle_array->data, sizeof(service_handle));
+
 	/* Now get service record */
 	dbus_g_proxy_begin_call (proxy,
 							"GetRemoteServiceRecord",
@@ -407,8 +423,21 @@ ods_bluez_get_client_socket (OdsBluez *bluez,
 	cb_data = g_new0 (GetClientSocketData, 1);
 	cb_data->cb = func;
 	cb_data->address = g_strdup (address);
-	cb_data->uuid = g_strdup (uuid);
 	cb_data->data = data;
+	/* From Johan Hedberg:
+	 *
+	 * some Nokia Symbian phones have two OBEX FTP services: one
+	 * identified with the normal UUID and another with a Nokia specific
+	 * 128 bit UUID. The service found behind the normal identifier is
+	 * very limited in features on these phones while the other one
+	 * supports full OBEX FTP (don't ask me why).
+	 */
+	/* if FTP was requested, use NOKIAFTP instead, 
+	 * if it isn't found we retreat to FTP in get_remote_service_handles_cb */
+	if (!strcmp (uuid, OBEX_FTP_UUID))
+		cb_data->uuid = g_strdup (OBEX_NOKIAFTP_UUID);
+	else
+		cb_data->uuid = g_strdup (uuid);
 	
 	/* Discover channel for needed service only if we don't know it yet */	
 	if (channel == 0) {
