@@ -1,6 +1,6 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
  *
- * Copyright (C) 2007 Tadas Dailyda <tadas@dailyda.com>
+ * Copyright (C) 2007-2008 Tadas Dailyda <tadas@dailyda.com>
  *
  * Licensed under the GNU General Public License Version 2
  *
@@ -37,6 +37,7 @@
 
 #include <glib.h>
 #include <dbus/dbus-glib.h>
+#include <dbus/dbus-glib-bindings.h>
 #include <dbus/dbus-glib-lowlevel.h>
 
 
@@ -100,13 +101,12 @@ G_DEFINE_TYPE (OdsManager, ods_manager, G_TYPE_OBJECT)
 
 static void     ods_manager_class_init	(OdsManagerClass *klass);
 static void     ods_manager_init		(OdsManager      *manager);
-static void		ods_manager_session_finalize (gpointer key,
+static gboolean ods_manager_session_finalize (gpointer key,
 										OdsManagerSessionInfo *session_info,
 										OdsManager *manager);
-static void		ods_manager_server_finalize (gpointer key,
+static gboolean		ods_manager_server_finalize (gpointer key,
 										OdsManagerServerInfo *server_info,
 										OdsManager *manager);
-static void     ods_manager_dispose		(GObject	 *object);
 static void     ods_manager_finalize	(GObject	 *object);
 
 
@@ -210,7 +210,8 @@ ods_manager_session_list_remove (OdsManager *manager, OdsSession *session)
 	g_object_get (session, "dbus-path", &session_object, NULL);
 	g_object_get (session, "owner", &owner, NULL);
 	
-	g_hash_table_remove (manager->priv->session_list, session_object);
+	if (!manager->priv->is_disposing)
+		g_hash_table_remove (manager->priv->session_list, session_object);
 	ods_manager_listened_names_remove (manager, owner, session_object);
 	
 	g_free (session_object);
@@ -226,7 +227,8 @@ ods_manager_server_list_remove (OdsManager *manager, OdsServer *server)
 	g_object_get (server, "dbus-path", &server_object, NULL);
 	g_object_get (server, "owner", &owner, NULL);
 	
-	g_hash_table_remove (manager->priv->server_list, server_object);
+	if (!manager->priv->is_disposing)
+		g_hash_table_remove (manager->priv->server_list, server_object);
 	ods_manager_listened_names_remove (manager, owner, server_object);
 	
 	g_free (server_object);
@@ -308,12 +310,16 @@ server_stopped_cb (OdsServer *server, OdsManager *manager)
 	
 	/* Remove SDP record (Bluetooth specific) */
 	server_info = g_hash_table_lookup (manager->priv->server_list, server_object);
-	if (strlen (server_info->bluetooth_address) > 0) {
+	if (!server_info)
+		goto out;
+
+	if (server_info->sdp_record_handle && strlen (server_info->bluetooth_address) > 0) {
 		ods_bluez_remove_service_record (manager->priv->bluez,
 											server_info->bluetooth_address,
 											server_info->sdp_record_handle);
 	}
 	
+out:	
 	g_free (server_object);
 }
 
@@ -370,7 +376,7 @@ dbus_name_owner_changed_cb (DBusGProxy *dbus_proxy, const gchar *name,
 	g_hash_table_insert (manager->priv->removed_dbus_names, g_strdup (name), manager);
 	
 	/* Now we finalize all objects (sessions and servers) */
-	object_list_dup = g_hash_table_get_keys (object_list);
+	object_list_dup = ods_hash_table_get_keys (object_list);
 	g_list_foreach (object_list_dup, (GFunc) ods_manager_finalize_object, manager);
 	g_list_free (object_list_dup);
 }
@@ -379,9 +385,11 @@ static void
 client_socket_connected_cb (gint fd, GError *error,
 							OdsManagerCreateBluetoothSessionData *data)
 {
-	OdsSession	*session;
-	gchar		*sender = NULL;
-	gchar		*session_object = NULL;
+	gboolean		name_exists = TRUE;
+	GError			*error_local = NULL;
+	OdsSession		*session;
+	gchar			*sender = NULL;
+	gchar			*session_object = NULL;
 	
 	
 	if (fd == -1) {
@@ -391,9 +399,27 @@ client_socket_connected_cb (gint fd, GError *error,
 		g_free (data);
 		return;
 	}
-	/* create session object and return it's object path */
 	sender = dbus_g_method_get_sender (data->context);
 	g_message ("Session created by: %s", sender);
+	/* It might be that sender disconnected from bus while connecting socket */
+	org_freedesktop_DBus_name_has_owner (data->manager->priv->dbus_proxy,
+											sender, &name_exists, &error_local);
+	if (error_local) {
+		/* ignore this */
+		g_clear_error (&error_local);
+	}
+	if (!name_exists) {
+		shutdown (fd, SHUT_RDWR);
+		close (fd);
+		g_message ("Session owner disconnected from bus, discarding session");
+		g_set_error (&error_local, ODS_ERROR, ODS_ERROR_FAILED,
+						"Session owner disconnected from bus");
+		dbus_g_method_return_error (data->context, error_local);
+		g_free (data->address);
+		g_free (data);
+		return;
+	}
+	/* create session object and return it's object path */
 	session = ods_session_new (fd, data->service, sender);
 	
 	/* add session to session list */
@@ -671,7 +697,6 @@ ods_manager_class_init (OdsManagerClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-	object_class->dispose = ods_manager_dispose;
 	object_class->finalize = ods_manager_finalize;
 
 	signals [SESSION_CREATED] =
@@ -762,7 +787,7 @@ ods_manager_init (OdsManager *manager)
 							G_OBJECT (manager));
 }
 
-static void
+static gboolean
 ods_manager_session_finalize (gpointer key, OdsManagerSessionInfo *session_info,
 								OdsManager *manager)
 {
@@ -773,9 +798,11 @@ ods_manager_session_finalize (gpointer key, OdsManagerSessionInfo *session_info,
 	/* Even if there was nothing to cancel, we will get
 	 * CANCELLED signal and disconnection will happen in 
 	 * session_cancelled_cb */
+
+	return TRUE;
 }
 
-static void
+static gboolean
 ods_manager_server_finalize (gpointer key, OdsManagerServerInfo *server_info,
 								OdsManager *manager)
 {
@@ -785,18 +812,16 @@ ods_manager_server_finalize (gpointer key, OdsManagerServerInfo *server_info,
 	server_stopped_cb (server, manager);
 	
 	g_signal_connect (server, "disposed", G_CALLBACK (server_disposed_cb), manager);
-	g_object_run_dispose (G_OBJECT (server));
+	ods_server_dispose (server);
+
+	return TRUE;
 }
 
-static void
-ods_manager_dispose (GObject *object)
+void
+ods_manager_dispose (OdsManager	*manager)
 {
-	OdsManager	*manager;
-
-	g_return_if_fail (object != NULL);
-	g_return_if_fail (ODS_IS_MANAGER (object));
-
-	manager = ODS_MANAGER (object);
+	g_return_if_fail (manager != NULL);
+	g_return_if_fail (ODS_IS_MANAGER (manager));
 
 	g_return_if_fail (manager->priv != NULL);
 	if (manager->priv->disposed)
@@ -810,13 +835,11 @@ ods_manager_dispose (GObject *object)
 		manager->priv->disposed = TRUE;
 		g_signal_emit (manager, signals [DISPOSED], 0);
 	} else {
-		g_hash_table_foreach (manager->priv->session_list,
-								(GHFunc) ods_manager_session_finalize, NULL);
-		g_hash_table_foreach (manager->priv->server_list,
-								(GHFunc) ods_manager_server_finalize, manager);
+		g_hash_table_foreach_remove (manager->priv->session_list,
+								(GHRFunc) ods_manager_session_finalize, manager);
+		g_hash_table_foreach_remove (manager->priv->server_list,
+								(GHRFunc) ods_manager_server_finalize, manager);
 	}
-	
-	G_OBJECT_CLASS (ods_manager_parent_class)->dispose (object);
 }
 
 /**
