@@ -1,6 +1,6 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
  *
- * Copyright (C) 2007 Tadas Dailyda <tadas@dailyda.com>
+ * Copyright (C) 2007-2008 Tadas Dailyda <tadas@dailyda.com>
  * Parts of code taken from osso-gwobex library by:
  * 				Johan Hedberg <johan.hedberg@nokia.com>
  *
@@ -92,11 +92,11 @@ ods_obex_get_new_path (const gchar *folder, const gchar *filename)
 {
 	gchar *path;
 	gchar *new_path;
-	gint iterator = 1;
+	guint iterator = 1;
 	gchar *first_part;
 	gchar *extension;
 	gchar *parentess;
-	gint pos;
+	guint pos;
 	
 	/* In case we don't know what the filename is (HDR_NAME wasn't received) */
 	if (filename == NULL || *filename == '\0')
@@ -137,6 +137,7 @@ ods_obex_transfer_new (OdsObexContext *obex_context, const gchar *local,
 	obex_context->modtime = -1;
 	obex_context->report_progress = TRUE; /* by default */
 	obex_context->transfer_started_signal_emitted = FALSE;
+	obex_context->suspended_once = FALSE;
 	obex_context->buf_size = 0;
 	obex_context->buf = NULL;
 	obex_context->stream_fd = -1;
@@ -285,7 +286,8 @@ ods_obex_connect (OdsObexContext *obex_context, const guchar *uuid,
 }
 
 gint
-ods_obex_srv_connect (OdsObexContext *obex_context, obex_object_t *object)
+ods_obex_srv_connect (OdsObexContext *obex_context, obex_object_t *object,
+							guint service)
 {
 	obex_headerdata_t	hv;
 	uint8_t				hi;
@@ -321,33 +323,65 @@ ods_obex_srv_connect (OdsObexContext *obex_context, obex_object_t *object)
 			target_len = hlen;
 		}
 	}
-	
+
 	OBEX_ObjectReParseHeaders (obex_context->obex_handle, object);
-	
-	if (target) {
-		hd.bs = target;
-		ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
-									OBEX_HDR_WHO, hd, target_len, 
-									OBEX_FL_FIT_ONE_PACKET);
-		if (ret < 0) {
-			OBEX_ObjectDelete (obex_context->obex_handle, object);
-			return ret;
-		}
-		
-		hd.bs = NULL;
-		hd.bq4 = 1; /* Connection ID is always 1 */
-		obex_context->connection_id = 1;
-		ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
-									OBEX_HDR_CONNECTION, hd, 4,
-									OBEX_FL_FIT_ONE_PACKET);
-		if (ret < 0) {
-			OBEX_ObjectDelete (obex_context->obex_handle, object);
-			return ret;
-		}
+
+	switch (service) {
+		case ODS_SERVICE_FTP:
+			/* Target header must be F9EC7BC4-953C-11D2-984E-525400DC9E09*/
+			if (!target || memcmp(OBEX_FTP_UUID, target, hlen) != 0) {
+				g_message("Target header Incorrect");
+				goto fail;
+			}
+			break;
+		case ODS_SERVICE_PBAP:
+			/* Target header must be 796135f0-f0c5-11d8-0966-0800200c9a66 */
+			if (!target || memcmp(OBEX_PBAP_UUID, target, hlen) != 0) {
+				g_message("Target header Incorrect");
+				goto fail;
+			}
+			break;
+		case ODS_SERVICE_OPP:
+			/* Target header must not be used */
+			if (target) {
+				g_message("Target header must not be used");
+				goto fail;
+			}
+			OBEX_ObjectSetRsp (object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
+
+			return 0;
+		default:
+			goto fail;
 	}
-	
+
+	hd.bs = target;
+	ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
+								OBEX_HDR_WHO, hd, target_len, 
+								OBEX_FL_FIT_ONE_PACKET);
+	if (ret < 0) {
+		OBEX_ObjectDelete (obex_context->obex_handle, object);
+		return ret;
+	}
+
+	hd.bs = NULL;
+	hd.bq4 = 1; /* Connection ID is always 1 */
+	obex_context->connection_id = 1;
+	ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
+								OBEX_HDR_CONNECTION, hd, 4,
+								OBEX_FL_FIT_ONE_PACKET);
+	if (ret < 0) {
+		OBEX_ObjectDelete (obex_context->obex_handle, object);
+		return ret;
+	}
+
 	OBEX_ObjectSetRsp (object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
-	
+
+	return 0;
+
+fail:
+
+	OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+
 	return 0;
 }
 
@@ -522,8 +556,8 @@ ods_obex_get (OdsObexContext *obex_context,
 	gint				ret;
 	obex_headerdata_t	hv;
 	obex_object_t		*object;
-	gunichar2			*uname;
-	glong				uname_len = 0;
+	gchar				*uname;
+	gsize				uname_len = 0;
 
 	g_assert (remote || type);
 
@@ -558,14 +592,14 @@ ods_obex_get (OdsObexContext *obex_context,
 	if (remote) {
 		
 
-		uname_len = ods_get_uname (&uname, remote);
-		if (uname_len < 0) {
+		uname_len = ods_filename_to_utf16 (&uname, remote);
+		if (uname == NULL) {
 			ret = -EINVAL;
 			goto out;
 		}
 
 		/* OpenOBEX is buggy and won't append the header unless hv.bs != NULL */
-		hv.bs = (guchar *) (uname ? (gchar *)uname : "");
+		hv.bs = (guchar *) uname;
 
 		ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object,
 									OBEX_HDR_NAME, hv, uname_len, 0);
@@ -620,7 +654,7 @@ ods_obex_srv_get (OdsObexContext *obex_context, obex_object_t *object,
 					/* This is GET by Type, leave remote and local = NULL */
 					break;
 				}
-				obex_context->remote = ods_utf16_to_utf8 (hv.bs, hlen);
+				obex_context->remote = ods_filename_from_utf16 ((gchar *) hv.bs, hlen);
 				obex_context->local = g_build_filename (current_path,
 														obex_context->remote,
 														NULL);
@@ -628,22 +662,43 @@ ods_obex_srv_get (OdsObexContext *obex_context, obex_object_t *object,
 				/* Check if such file exists */
 				if (!g_file_test (obex_context->local, G_FILE_TEST_EXISTS)) {
 					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
+					goto out;
+				}
+				if (!g_file_test (obex_context->local, G_FILE_TEST_IS_REGULAR)) {
+					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+					goto out;
+				}
+				if (g_access (obex_context->local, R_OK) < 0) {
+					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_UNAUTHORIZED, OBEX_RSP_UNAUTHORIZED);
 					goto out;
 				}
 				break;
 				
 			case OBEX_HDR_TYPE:
-				/* Ensure null termination */
-				if (hv.bs[hlen - 1] != '\0')
-					break;
-				if (!g_utf8_validate ((const gchar *) hv.bs, -1, NULL)) {
-					//debug("Invalid type header: %s", hv.bs);
-					break;
+				if (hv.bs[hlen - 1] != '\0' ||
+						!g_utf8_validate ((const gchar *) hv.bs, -1, NULL)) {
+					/* invalid type header */
+					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
+					goto out;
 				}
 				
 				obex_context->type = g_strdup ((const gchar *) hv.bs);
+				g_message ("HDR_TYPE: %s", obex_context->type);
+				break;
+			
+			case OBEX_HDR_CONNECTION:
+				if (obex_context->connection_id != CONID_INVALID &&
+						hv.bq4 != obex_context->connection_id) {
+					/* wrong connection id */
+					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
+					goto out;
+				}
 				break;
-				
 			default:
 				break;
 		}
@@ -658,6 +713,7 @@ ods_obex_srv_get (OdsObexContext *obex_context, obex_object_t *object,
 		obex_context->stream_fd = open (obex_context->local, O_RDONLY);
 		if (obex_context->stream_fd < 0) {
 			ret = -1;
+			OBEX_ObjectSetRsp (object, OBEX_RSP_INTERNAL_SERVER_ERROR, OBEX_RSP_INTERNAL_SERVER_ERROR);
 			goto out;
 		}
 		/* allocate buffer */
@@ -683,8 +739,12 @@ ods_obex_srv_get (OdsObexContext *obex_context, obex_object_t *object,
 				hv.bs = (guchar *)tstr;
 				ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
 											OBEX_HDR_TIME, hv, len, 0);
-				if (ret < 0)
+				if (ret < 0) {
+					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_INTERNAL_SERVER_ERROR, OBEX_RSP_INTERNAL_SERVER_ERROR);
 					goto out;
+				}
+				
 			}
 		}
 
@@ -694,8 +754,11 @@ ods_obex_srv_get (OdsObexContext *obex_context, obex_object_t *object,
 			hv.bq4 = (uint32_t)object_size;
 			ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
 										OBEX_HDR_LENGTH, hv, 4, 0);
-			if (ret < 0)
+			if (ret < 0) {
+				ret = -1;
+				OBEX_ObjectSetRsp (object, OBEX_RSP_INTERNAL_SERVER_ERROR, OBEX_RSP_INTERNAL_SERVER_ERROR);
 				goto out;
+			}
 		} else {
 			obex_context->target_size = 0;
 		}
@@ -714,11 +777,13 @@ ods_obex_srv_get (OdsObexContext *obex_context, obex_object_t *object,
 		} else {
 			/* currently no other types are supported */
 			ret = -1;
+			OBEX_ObjectSetRsp (object, OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
 			goto out;
 		}
 	} else {
 		/* neither name nor type was specified */
 		ret = -1;
+		OBEX_ObjectSetRsp (object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
 		goto out;
 	}
 	
@@ -727,10 +792,9 @@ ods_obex_srv_get (OdsObexContext *obex_context, obex_object_t *object,
 	ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
 								OBEX_HDR_BODY, hv, 0, OBEX_FL_STREAM_START);
 out:
-	if (ret < 0) {
-		OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+	if (ret < 0)
 		ods_obex_transfer_close (obex_context);
-	} else
+	else
 		OBEX_ObjectSetRsp (object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
 	return ret;
 }
@@ -743,8 +807,8 @@ ods_obex_put (OdsObexContext *obex_context,
 	gint				ret;
 	obex_headerdata_t	hv;
 	obex_object_t		*object = NULL;
-	gunichar2			*uname = NULL;
-	glong				uname_len = 0;
+	gchar				*uname = NULL;
+	gsize				uname_len = 0;
 	gint				object_size = 0;
 	time_t				object_time = -1;
 
@@ -755,8 +819,8 @@ ods_obex_put (OdsObexContext *obex_context,
 	
 	/* get UTF name for remote file */
 	if (remote) {
-		uname_len = ods_get_uname (&uname, remote);
-		if (uname_len < 0) {
+		uname_len = ods_filename_to_utf16 (&uname, remote);
+		if (uname == NULL) {
 			ret = -EINVAL;
 			goto out;
 		}
@@ -797,7 +861,7 @@ ods_obex_put (OdsObexContext *obex_context,
 
 	/* Add name header */
 	if (uname) {
-		hv.bs = (guchar *)uname;
+		hv.bs = (guchar *) uname;
 		ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
 									OBEX_HDR_NAME, hv, uname_len, 0);
 		if (ret < 0)
@@ -862,7 +926,7 @@ ods_obex_put (OdsObexContext *obex_context,
 
 	ret = ods_obex_send (obex_context->obex_handle, object);
 out:
-	if (uname_len >= 0)
+	if (uname_len > 0)
 		g_free (uname);
 	if (ret < 0 && object)
 		OBEX_ObjectDelete (obex_context->obex_handle, object);
@@ -880,8 +944,9 @@ ods_obex_srv_put (OdsObexContext *obex_context, obex_object_t *object,
 	uint8_t				hi;
 	guint				hlen;
 	gint				ret = 0;
-	gint				written = 0;
+	guint				written = 0;
 	gint				write_ret;
+	gboolean			is_delete = TRUE;
 	
 	/* Check if we already have all transfer info
 	 * because both EV_REQCHECK and EV_REQ trigger this function */
@@ -893,21 +958,22 @@ ods_obex_srv_put (OdsObexContext *obex_context, obex_object_t *object,
 		g_message ("header: %d", hi);
 		switch (hi) {
 			case OBEX_HDR_BODY:
+				is_delete = FALSE;
 				g_message ("HDR_BODY length=%u", hlen);
 				break;
 				
 			case OBEX_HDR_NAME:
-				obex_context->remote = ods_utf16_to_utf8 (hv.bs, hlen);
+				obex_context->remote = ods_filename_from_utf16 ((gchar *) hv.bs, hlen);
 				g_message ("HDR_NAME: %s", obex_context->remote);
 				break;
 				
 			case OBEX_HDR_TYPE:
-				/* Ensure null termination */
-				if (hv.bs[hlen - 1] != '\0')
-					break;
-				if (!g_utf8_validate ((const gchar *) hv.bs, -1, NULL)) {
-					//debug("Invalid type header: %s", hv.bs);
-					break;
+				if (hv.bs[hlen - 1] != '\0' ||
+					!g_utf8_validate ((const gchar *) hv.bs, -1, NULL)) {
+					/* invalid type header */
+					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
+					goto out;
 				}
 				
 				obex_context->type = g_strdup ((const gchar *) hv.bs);
@@ -916,11 +982,13 @@ ods_obex_srv_put (OdsObexContext *obex_context, obex_object_t *object,
 				
 			case OBEX_HDR_LENGTH:
 				obex_context->target_size = hv.bq4;
+				is_delete = FALSE;
 				g_message ("HDR_LENGTH: %" G_GUINT64_FORMAT, obex_context->target_size);
 				break;
 				
 			case OBEX_HDR_TIME:
 				obex_context->modtime = ods_parse_iso8601 ((gchar*) hv.bs, hlen);
+				g_message ("HDR_TIME");
 				break;
 				
 			case OBEX_HDR_DESCRIPTION:
@@ -928,10 +996,20 @@ ods_obex_srv_put (OdsObexContext *obex_context, obex_object_t *object,
 				break;
 				
 			case OBEX_HDR_COUNT:
-				/* This informs us how many objects is client going to send
+				/* This informs us how many objects client is going to send
 				 * during this session. We really don't care. */
 				break;
-				
+			
+			case OBEX_HDR_CONNECTION:
+				if (obex_context->connection_id != CONID_INVALID &&
+						hv.bq4 != obex_context->connection_id) {
+					/* wrong connection id */
+					ret = -1;
+					OBEX_ObjectSetRsp (object, OBEX_RSP_BAD_REQUEST, OBEX_RSP_BAD_REQUEST);
+					goto out;
+				}
+				break;
+
 			default:
 				break;
 		}
@@ -946,10 +1024,34 @@ ods_obex_srv_put (OdsObexContext *obex_context, obex_object_t *object,
 		obex_context->report_progress = TRUE;
 		if (!ods_obex_srv_new_file (obex_context, path)) {
 			ret = -1;
+			OBEX_ObjectSetRsp (object, OBEX_RSP_UNAUTHORIZED, OBEX_RSP_UNAUTHORIZED);
 			goto out;
 		}
+	} else if (is_delete) {
+		/* this is a delete request */
+		obex_context->local = g_build_filename (path, obex_context->remote, NULL);
+
+		if (!g_file_test (obex_context->local, G_FILE_TEST_EXISTS)) {
+			ret = -1;
+			OBEX_ObjectSetRsp (object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
+			goto out;
+		}
+		if (g_access (obex_context->local, W_OK) < 0) {
+			ret = -1;
+			OBEX_ObjectSetRsp (object, OBEX_RSP_UNAUTHORIZED, OBEX_RSP_UNAUTHORIZED);
+			goto out;
+		}
+		
+		g_message ("Deleting: %s", obex_context->local);
+		if (g_file_test (obex_context->local, G_FILE_TEST_IS_DIR))
+			ret = rmdir (obex_context->local);
+		else
+			ret = g_unlink (obex_context->local);
+						
+		if (ret == -1)
+			OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 	}
-	
+
 	/* If there is some data received already, write it to file */
 	while (written < obex_context->buf_size) {
 		write_ret = write (obex_context->stream_fd, obex_context->buf + written, 
@@ -959,6 +1061,7 @@ ods_obex_srv_put (OdsObexContext *obex_context, obex_object_t *object,
 
 		if (write_ret < 0) {
 			ret = -errno;
+			OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 			break;
 		}
 
@@ -966,12 +1069,9 @@ ods_obex_srv_put (OdsObexContext *obex_context, obex_object_t *object,
 	}
 
 out:
-	if (ret < 0) {
-		OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
-		/* either delete operation was requested but not allowed
-		 * or operation failed */
+	if (ret < 0)
 		ods_obex_transfer_close (obex_context);
-	} else
+	else
 		OBEX_ObjectSetRsp (object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
 	return ret;
 }
@@ -984,8 +1084,8 @@ ods_obex_setpath (OdsObexContext *obex_context, const gchar *path,
 	obex_headerdata_t	hv;
 	obex_object_t		*object = NULL;
 	obex_setpath_hdr_t	nonhdrdata;
-	gunichar2			*uname; 
-	glong				uname_len;
+	gchar				*uname; 
+	gsize				uname_len;
 
 	nonhdrdata.flags = 0x02;
 	nonhdrdata.constants = 0;
@@ -993,11 +1093,11 @@ ods_obex_setpath (OdsObexContext *obex_context, const gchar *path,
 	if (strcmp (path, "..") == 0) {
 		/* move up one directory */
 		nonhdrdata.flags = 0x03;
-		uname_len = -1;
+		uname_len = 0;
 	} else {
 		/* normal directory change */
-		uname_len = ods_get_uname (&uname, path);
- 		if (uname_len < 0) {
+		uname_len = ods_filename_to_utf16 (&uname, path);
+ 		if (uname == NULL) {
  			ret = -EINVAL;
 			goto out;
  		}
@@ -1028,17 +1128,15 @@ ods_obex_setpath (OdsObexContext *obex_context, const gchar *path,
 	}
 
 	/* Add name header */
-	if (uname_len >= 0) {
-		hv.bs = (guchar *) (uname ? (gchar *)uname : "");
-        ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
-        							OBEX_HDR_NAME, hv, uname_len, 0);
-        if (ret < 0)
-        	goto out;
-    }
+	hv.bs = (guchar *) uname;
+    ret = OBEX_ObjectAddHeader (obex_context->obex_handle, object, 
+    							OBEX_HDR_NAME, hv, uname_len, 0);
+    if (ret < 0)
+    	goto out;
 
 	ret = ods_obex_send (obex_context->obex_handle, object);
 out:
-	if (uname_len >= 0 )
+	if (uname_len > 0 )
 		g_free (uname);
 	if (ret < 0 && object)
 		OBEX_ObjectDelete (obex_context->obex_handle, object);
@@ -1093,7 +1191,7 @@ ods_obex_srv_setpath (OdsObexContext *obex_context, obex_object_t *object,
 		if (hi == OBEX_HDR_NAME) {
 			if (hlen > 0) {
 				/* Normal directory change */
-				directory = ods_utf16_to_utf8 (hv.bs, hlen);
+				directory = ods_filename_from_utf16 ((gchar *) hv.bs, hlen);
 				*new_path = g_build_filename (current_path, directory, NULL);
 				g_free (directory);
 				/* Check if such path exists */
