@@ -1,6 +1,6 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
  *
- * Copyright (C) 2007 Tadas Dailyda <tadas@dailyda.com>
+ * Copyright (C) 2007-2008 Tadas Dailyda <tadas@dailyda.com>
  *
  * Licensed under the GNU General Public License Version 2
  *
@@ -74,8 +74,7 @@ struct OdsServerSessionPrivate
 	OdsServerSessionState	state; /* ODS_SERVER_SESSION_STATE_OPEN by default */
 	/* OBEX connection */
 	OdsObexContext			*obex_context;
-	GIOChannel				*io_channel;
-	GSource					*io_source;
+	guint					io_watch;
 	/* other */
 	GStaticMutex			mutex;
 	DBusGMethodInvocation	*dbus_context; /* D-Bus context for async methods */
@@ -98,6 +97,15 @@ static guint	iterator = 0;
 
 G_DEFINE_TYPE (OdsServerSession, ods_server_session, G_TYPE_OBJECT)
 
+/* This is to be called from mainloop, not directly. It's to ensure
+ * that we don't finalize this object from within itself. */
+static gboolean
+emit_disconnected (OdsServerSession *server_session)
+{
+	g_signal_emit (server_session, signals [DISCONNECTED], 0);
+	return FALSE;
+}
+
 static gboolean
 obex_io_callback (GIOChannel *io_channel, GIOCondition cond, gpointer data)
 {
@@ -133,7 +141,7 @@ obex_io_callback (GIOChannel *io_channel, GIOCondition cond, gpointer data)
 	}
 	if (!ret) {
 		/* Also emit DISCONNECTED signal, since this session is now defunct */
-		g_signal_emit (server_session, signals [DISCONNECTED], 0);
+		g_idle_add ((GSourceFunc) emit_disconnected, server_session);
 	}
 
 	return ret;
@@ -218,21 +226,6 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 					ods_server_session_disconnect_internal (server_session);
 					break;
 				case OBEX_CMD_PUT:
-					if (obex_context->local == NULL) {
-						/* This is a Delete request */
-						obex_context->local = g_build_filename (
-													server_session->priv->path, 
-													obex_context->remote, NULL);
-						g_message ("Deleting: %s", obex_context->local);
-						if (g_file_test (obex_context->local, G_FILE_TEST_IS_DIR))
-							ret = rmdir (obex_context->local);
-						else
-							ret = g_unlink (obex_context->local);
-						
-						if (ret == -1)
-							OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN,
-												OBEX_RSP_FORBIDDEN);
-					}
 					/* Transfer complete */
 					ods_obex_transfer_close (obex_context);
 					server_session->priv->state = ODS_SERVER_SESSION_STATE_OPEN;
@@ -253,12 +246,6 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 		case OBEX_EV_REQHINT:
 			switch (command) {
 				case OBEX_CMD_PUT:
-					if (!server_session->priv->allow_write) {
-						g_message ("CMD_PUT forbidden");
-						OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, 
-											OBEX_RSP_FORBIDDEN);
-						return;
-					}
 					OBEX_ObjectReadStream (handle, object, NULL);
 					obex_context->obex_cmd = OBEX_CMD_PUT;
 					/* Initialize transfer and set state */
@@ -299,6 +286,13 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 										OBEX_RSP_FORBIDDEN);
 					return;
 				}
+				/* Check if we have write permissions for current path */
+				if (g_access (server_session->priv->path, W_OK) < 0) {
+					g_message ("No write permissions for current path");
+					OBEX_ObjectSetRsp (object, OBEX_RSP_UNAUTHORIZED, 
+										OBEX_RSP_UNAUTHORIZED);
+					return;
+				}
 				ret = ods_obex_srv_put (obex_context, object, 
 										server_session->priv->path);
 				g_message ("ret=%d", ret);
@@ -310,8 +304,6 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 									obex_context->target_size);
 					obex_context->transfer_started_signal_emitted = TRUE;
 				}
-				if (!server_session->priv->auto_accept && ret == 0)
-					OBEX_SuspendRequest (obex_context->obex_handle, object);
 			}
 			break;
 		case OBEX_EV_REQ:
@@ -322,10 +314,15 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 				             
 				case OBEX_CMD_CONNECT:
 					g_message ("CMD_CONNECT requested");
-					ods_obex_srv_connect (obex_context, object);
+					ods_obex_srv_connect (obex_context, object, server_session->priv->service);
 					break;
 				          
 				case OBEX_CMD_SETPATH:
+					if (server_session->priv->service == ODS_SERVICE_OPP){
+						OBEX_ObjectSetRsp (object, OBEX_RSP_NOT_IMPLEMENTED, 
+											OBEX_RSP_NOT_IMPLEMENTED);
+						return;
+					}
 					g_message ("CMD_SETPATH requested");
 					g_message ("current path: %s", server_session->priv->path);
 					g_message ("root path: %s", server_session->priv->root_path);
@@ -361,6 +358,13 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 											OBEX_RSP_FORBIDDEN);
 						return;
 					}
+					/* Check if we have write permissions for current path */
+					if (g_access (server_session->priv->path, W_OK) < 0) {
+						g_message ("No write permissions for current path");
+						OBEX_ObjectSetRsp (object, OBEX_RSP_UNAUTHORIZED, 
+											OBEX_RSP_UNAUTHORIZED);
+						return;
+					}
 					ret = ods_obex_srv_put (obex_context, object, 
 											server_session->priv->path);
 					g_message ("ret=%d", ret);
@@ -371,8 +375,6 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 										obex_context->target_size);
 						obex_context->transfer_started_signal_emitted = TRUE;
 					}
-					if (!server_session->priv->auto_accept && ret == 0)
-						OBEX_SuspendRequest (obex_context->obex_handle, object);
 					break;
 				default:
 					OBEX_ObjectSetRsp (object, OBEX_RSP_NOT_IMPLEMENTED, 
@@ -385,6 +387,21 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 			obex_transfer_data_exchange_done (server_session, ret);
 			break;
 		case OBEX_EV_STREAMAVAIL:
+			if (!server_session->priv->allow_write) {
+				g_message ("CMD_PUT forbidden");
+				OBEX_ObjectSetRsp (object, OBEX_RSP_FORBIDDEN, 
+									OBEX_RSP_FORBIDDEN);
+				return;
+			}
+			/* Check if we have write permissions for current path */
+			if (g_access (server_session->priv->path, W_OK) < 0) {
+				g_message ("No write permissions for current path");
+				OBEX_ObjectSetRsp (object, OBEX_RSP_UNAUTHORIZED, 
+									OBEX_RSP_UNAUTHORIZED);
+				return;
+			}
+			ret = 0;
+						
 			/* This PUT request is definitely not a delete request. Let's
 			 * open a file for writing. */
 			if (obex_context->remote && !obex_context->local) {
@@ -392,7 +409,15 @@ obex_event (obex_t *handle, obex_object_t *object, int mode, int event,
 				if (!ods_obex_srv_new_file (obex_context, server_session->priv->path))
 					ret = -1;
 			}
-			ret = ods_obex_readstream (obex_context, object);
+			if (ret == 0)
+				ret = ods_obex_readstream (obex_context, object);
+			/* only suspend if we have transfer info */
+			if (!server_session->priv->auto_accept && obex_context->local &&
+					!obex_context->suspended_once) {
+				g_message ("suspending request");
+				obex_context->suspended_once = TRUE;
+				OBEX_SuspendRequest (obex_context->obex_handle, object);
+			}
 			obex_transfer_data_exchange_done (server_session, ret);
 			break;
 		case OBEX_EV_PARSEERR:
@@ -653,6 +678,7 @@ ods_server_session_constructor (GType type, guint n_construct_params,
 	GObject *object;
 	OdsServerSession *server_session;
 	OdsObexContext *obex_context;
+	GIOChannel *chan;
 	gint ret;
 	/*GError *error = NULL;*/
 	g_message ("Creating server session");
@@ -687,18 +713,12 @@ ods_server_session_constructor (GType type, guint n_construct_params,
 		g_warning ("error transport setup fail");
 		return NULL;
 	}
-
-	server_session->priv->io_channel = g_io_channel_unix_new (
-													server_session->priv->fd);
-	
-	server_session->priv->io_source = g_io_create_watch (
-									server_session->priv->io_channel,
-									G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL);
-    g_source_set_callback (server_session->priv->io_source, 
-							(GSourceFunc)obex_io_callback, 
-							obex_context->obex_handle, NULL);
-    (void) g_source_attach (server_session->priv->io_source, NULL);
-	g_source_unref (server_session->priv->io_source);
+	chan = g_io_channel_unix_new (server_session->priv->fd);
+	g_io_channel_set_close_on_unref (chan, TRUE);
+	server_session->priv->io_watch = g_io_add_watch(chan,
+							G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+							obex_io_callback, obex_context->obex_handle);
+	g_io_channel_unref(chan);
 	
 	return object;
 }
@@ -723,12 +743,11 @@ ods_server_session_finalize (GObject *object)
 
 	/* close connection, free obex_context */
 	g_message ("closing connection");
-	OBEX_TransportDisconnect (server_session->priv->obex_context->obex_handle);
+	shutdown (server_session->priv->fd, SHUT_RDWR);
+	close (server_session->priv->fd);
 	OBEX_Cleanup (server_session->priv->obex_context->obex_handle);
 	g_free (server_session->priv->obex_context);
-	g_io_channel_shutdown (server_session->priv->io_channel, FALSE, NULL);
-	g_io_channel_unref (server_session->priv->io_channel);
-	g_source_destroy (server_session->priv->io_source);
+
 	/* free other private variables */
 	g_free (server_session->priv->root_path);
 	g_free (server_session->priv->path);
@@ -826,7 +845,8 @@ ods_server_session_disconnect_internal (OdsServerSession *server_session)
 		OBEX_TransportDisconnect (server_session->priv->obex_context->obex_handle);
 		server_session->priv->state = ODS_SERVER_SESSION_STATE_NOT_CONNECTED;
 	}
-	g_signal_emit (server_session, signals [DISCONNECTED], 0);
+	g_source_remove (server_session->priv->io_watch);
+	g_idle_add ((GSourceFunc) emit_disconnected, server_session);
 }
 
 gboolean
