@@ -1,6 +1,6 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
  *
- * Copyright (C) 2007 Tadas Dailyda <tadas@dailyda.com>
+ * Copyright (C) 2007-2008 Tadas Dailyda <tadas@dailyda.com>
  *
  * Licensed under the GNU General Public License Version 2
  *
@@ -55,7 +55,6 @@ static void     ods_server_class_init	(OdsServerClass *klass);
 static void     ods_server_init		(OdsServer     *server);
 static GObject* ods_server_constructor (GType type, guint n_construct_params, 
 									GObjectConstructParam *construct_params);
-static void		ods_server_dispose (GObject		*object);
 static void     ods_server_finalize	(GObject		*object);
 
 #define ODS_SERVER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), ODS_TYPE_SERVER, OdsServerPrivate))
@@ -68,7 +67,11 @@ typedef struct OdsServerSessionInfo_ {
 
 struct OdsServerPrivate
 {
+	/* viariables used when disposing/stopping */
 	gboolean				is_disposing;
+	gboolean				is_stopping;
+	guint					open_sessions;
+	gboolean				socket_error;
 	gboolean				disposed;
 	/* constructor properties */
 	gint					fd; /* rfcomm device */
@@ -135,17 +138,30 @@ server_session_disconnected_cb (OdsServerSession *server_session,
 	
 	g_message ("server session closed");
 	g_object_get (server_session, "dbus-path", &session_object, NULL);
-	g_hash_table_remove (server->priv->session_list, session_object);
+
+	if (!server->priv->is_disposing && !server->priv->is_stopping)
+		g_hash_table_remove (server->priv->session_list, session_object);
+	else
+		server->priv->open_sessions--;
+
 	g_signal_emit (server, signals [SESSION_REMOVED], 0, session_object);
 	
 	g_object_unref (server_session);
 	g_free (session_object);
 	
-	if (server->priv->is_disposing && 
-		g_hash_table_size (server->priv->session_list) == 0) {
+	if (server->priv->open_sessions == 0) {
+		if (server->priv->is_disposing) {
 			g_message ("Server disposed");
 			server->priv->disposed = TRUE;
 			g_signal_emit (server, signals [DISPOSED], 0);
+		} else if (server->priv->is_stopping) {
+			g_message ("Server stopped");
+			server->priv->is_stopping = FALSE;
+			server->priv->started = FALSE;
+			g_signal_emit (server, signals [STOPPED], 0);
+			if (server->priv->socket_error)
+				g_signal_emit (server, signals [CLOSED], 0);
+		}
 	}
 }
 
@@ -156,17 +172,19 @@ server_session_cancelled_cb (OdsServerSession *server_session)
 	ods_server_session_disconnect_internal (server_session);
 }
 
-static void
+static gboolean
 server_session_finalize (gpointer key, OdsServerSessionInfo *session_info,
 							OdsServer *server)
 {
 	g_message ("attempting to close server session");
 	g_signal_connect (session_info->session, "cancelled", 
 						G_CALLBACK (server_session_cancelled_cb), NULL);
-	ods_server_session_cancel_internal (session_info->session);
+	g_idle_add ((GSourceFunc) ods_server_session_cancel_internal, session_info->session);
 	/* Even if there was nothing to cancel, we will get
 	 * CANCELLED signal and disconnection will happen in 
 	 * session_cancelled_cb */
+
+	return TRUE;
 }
 
 static gboolean
@@ -246,13 +264,18 @@ out:
 	}
 	if (!ret) {
 		/* disconnect server sessions, cause there was error on server socket */
-		g_hash_table_foreach (server->priv->session_list,
-								(GHFunc) server_session_finalize, NULL);
+		if (g_hash_table_size (server->priv->session_list) == 0) {
+			server->priv->started = FALSE;
+			g_signal_emit (server, signals [STOPPED], 0);
+			g_signal_emit (server, signals [CLOSED], 0);
+		} else {
+			server->priv->socket_error = TRUE;
+			server->priv->is_stopping = TRUE;
+			g_hash_table_foreach_remove (server->priv->session_list,
+										(GHRFunc) server_session_finalize, NULL);
+			/* STOPPED and CLOSED signals will be emitted in server_session_disconnected_cb */
+		}
 		g_hash_table_unref (server->priv->session_list);
-		/* emit STOPPED and CLOSED */
-		server->priv->started = FALSE;
-		g_signal_emit (server, signals [STOPPED], 0);
-		g_signal_emit (server, signals [CLOSED], 0);
 	}
 	
     return ret;
@@ -324,7 +347,6 @@ ods_server_class_init (OdsServerClass *klass)
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 	
 	object_class->constructor = ods_server_constructor;
-	object_class->dispose = ods_server_dispose;
 	object_class->finalize = ods_server_finalize;
 	
 	object_class->set_property = ods_server_set_property;
@@ -508,38 +530,6 @@ ods_server_constructor (GType type, guint n_construct_params,
 	return object;
 }
 
-static void
-ods_server_dispose (GObject *object)
-{
-	OdsServer *server;
-
-	g_return_if_fail (object != NULL);
-	g_return_if_fail (ODS_IS_SERVER (object));
-
-	server = ODS_SERVER (object);
-
-	g_return_if_fail (server->priv != NULL);
-	if (server->priv->disposed)
-		return;
-	
-	g_message ("Disposing Server");
-	server->priv->is_disposing = TRUE;
-	/* check if there is nothing to dispose */
-	if (g_hash_table_size (server->priv->session_list) == 0) {
-		g_message ("Server disposed at once");
-		server->priv->disposed = TRUE;
-		g_signal_emit (server, signals [DISPOSED], 0);
-	} else {
-		/* disconnect server sessions */
-		g_hash_table_foreach (server->priv->session_list,
-								(GHFunc) server_session_finalize, NULL);
-		/* when session_list becomes empty in server_session_disconnected callback
-		 * "disposed" signal will be emitted */
-	}
-	
-	G_OBJECT_CLASS (ods_server_parent_class)->dispose (object);
-}
-
 /**
  * ods_server_finalize:
  * @object: The object to finalize
@@ -606,6 +596,13 @@ ods_server_start (OdsServer *server, const gchar *path, gboolean allow_write,
 		g_clear_error (&error);
 		return FALSE;
 	}
+	if (server->priv->is_stopping || server->priv->is_disposing) {
+		g_set_error (&error, ODS_ERROR, ODS_ERROR_FAILED,
+						"Currently stopping");
+		dbus_g_method_return_error (context, error);
+		g_clear_error (&error);
+		return FALSE;
+	}
 	/* make sure supplied path is valid */
 	if (!g_file_test (path, G_FILE_TEST_IS_DIR)) {
 		g_set_error (&error, ODS_ERROR, ODS_ERROR_INVALID_ARGUMENTS,
@@ -643,15 +640,19 @@ ods_server_stop (OdsServer *server, DBusGMethodInvocation *context)
 		return FALSE;
 	}
 	/* disconnect server sessions */
-	g_hash_table_foreach (server->priv->session_list,
-							(GHFunc) server_session_finalize, NULL);
+	if (g_hash_table_size (server->priv->session_list) == 0) {
+		server->priv->started = FALSE;
+		g_signal_emit (server, signals [STOPPED], 0);
+	} else {
+		server->priv->is_stopping = TRUE;
+		g_hash_table_foreach_remove (server->priv->session_list,
+									(GHRFunc) server_session_finalize, NULL);
+		/* STOPPED signal will be emitted in server_session_disconnected_cb */
+	}
 	g_hash_table_unref (server->priv->session_list);
 	server->priv->session_list = g_hash_table_new_full (g_str_hash, g_str_equal,
 								g_free, (GDestroyNotify) ods_server_session_info_free);
 	
-	/* Change state to stopped */
-	server->priv->started = FALSE;
-	g_signal_emit (server, signals [STOPPED], 0);
 	dbus_g_method_return (context);
 	return TRUE;
 }
@@ -670,6 +671,33 @@ ods_server_close (OdsServer *server, DBusGMethodInvocation *context)
 	return TRUE;
 }
 
+/* Not exposed through DBUS, used internally */
+void
+ods_server_dispose (OdsServer *server)
+{
+	g_return_if_fail (server != NULL);
+	g_return_if_fail (ODS_IS_SERVER (server));
+	g_return_if_fail (server->priv != NULL);
+	if (server->priv->disposed)
+		return;
+	
+	g_message ("Disposing Server");
+	server->priv->is_disposing = TRUE;
+	/* check if there is nothing to dispose */
+	if (g_hash_table_size (server->priv->session_list) == 0) {
+		g_message ("Server disposed at once");
+		server->priv->disposed = TRUE;
+		g_signal_emit (server, signals [DISPOSED], 0);
+	} else {
+		/* disconnect server sessions */
+		server->priv->open_sessions = g_hash_table_size (server->priv->session_list);
+		g_hash_table_foreach_remove (server->priv->session_list,
+								(GHRFunc) server_session_finalize, NULL);
+		/* when session_list becomes empty in server_session_disconnected callback
+		 * "disposed" signal will be emitted */
+	}
+}
+
 gboolean
 ods_server_is_started (OdsServer *server)
 {
