@@ -1,6 +1,6 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
+/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
  *
- * Copyright (C) 2007 Tadas Dailyda <tadas@dailyda.com>
+ * Copyright (C) 2007-2008 Tadas Dailyda <tadas@dailyda.com>
  *
  * Licensed under the GNU General Public License Version 2
  *
@@ -72,8 +72,7 @@ struct OdsSessionPrivate
 	OdsSessionState			state; /* ODS_SESSION_STATE_NOT_CONNECTED by default */
 	/* OBEX connection */
 	OdsObexContext			*obex_context;
-	GIOChannel				*io_channel;
-	GSource					*io_source;
+	guint					io_watch;
 	/* other */
 	GStaticMutex			mutex;
 	DBusGMethodInvocation	*dbus_context; /* D-Bus context for async methods */
@@ -102,6 +101,15 @@ static guint	iterator = 0;
 
 G_DEFINE_TYPE (OdsSession, ods_session, G_TYPE_OBJECT)
 
+/* This is to be called from mainloop, not directly. It's to ensure
+ * that we don't finalize this object from within itself. */
+static gboolean
+emit_disconnected (OdsSession *session)
+{
+	g_signal_emit (session, signals [DISCONNECTED], 0);
+	return FALSE;
+}
+
 static gboolean
 obex_io_callback (GIOChannel *io_channel, GIOCondition cond, gpointer data)
 {
@@ -115,6 +123,12 @@ obex_io_callback (GIOChannel *io_channel, GIOCondition cond, gpointer data)
 	
 	g_message ("io callback");
 	if (cond & (G_IO_ERR | G_IO_HUP | G_IO_NVAL)) {
+		g_source_remove (session->priv->io_watch);
+		g_idle_add ((GSourceFunc) emit_disconnected, session);
+		/* CMD_DISCONNECT was sent but did not receive proper reply,
+		 * don't emit ErrorOccurred */
+		if (session->priv->state == ODS_SESSION_STATE_NOT_CONNECTED)
+			return FALSE;
 		g_set_error (&error, ODS_ERROR, ODS_ERROR_LINK_ERROR, "Connection error");
 		/* cleanup transfer data and set state to NOT_CONNECTED */
 		/* If it was GET operation, remove incomplete file */
@@ -239,8 +253,8 @@ obex_request_done (OdsSession *session, obex_object_t *object, int command,
 			}
 			break;
 		case OBEX_CMD_DISCONNECT:
-			session->priv->state = ODS_SESSION_STATE_NOT_CONNECTED;
-			g_signal_emit (session, signals [DISCONNECTED], 0);
+			g_source_remove (session->priv->io_watch);
+			g_idle_add ((GSourceFunc) emit_disconnected, session);
 			break;
 		case OBEX_CMD_SETPATH:
 			/* check response code here */
@@ -266,9 +280,14 @@ obex_request_done (OdsSession *session, obex_object_t *object, int command,
 					/* add trailing "/" */
 					temp = g_strdup_printf ("%s/", temp2);
 					g_free (temp2);
-				}
-				else {
-					temp = g_strdup_printf ("%s%s/", session->priv->current_path,
+				} else {
+					gchar *first_element;
+					/* make sure we don't leave "/" in the beginning */
+					if (session->priv->current_path[0] == '/')
+						first_element = "";
+					else
+						first_element = session->priv->current_path;
+					temp = g_strdup_printf ("%s%s/", first_element,
 												session->priv->new_path);
 				}
 				g_free (session->priv->current_path);
@@ -428,6 +447,7 @@ static gboolean
 ods_session_setup_transport (OdsSession *session)
 {
 	OdsObexContext *obex_context;
+	GIOChannel *chan;
 	gint ret;
 	GError *error = NULL;
 	
@@ -458,15 +478,12 @@ ods_session_setup_transport (OdsSession *session)
 		return FALSE;
 	}
 
-	session->priv->io_channel = g_io_channel_unix_new (session->priv->fd);
-	
-	session->priv->io_source = g_io_create_watch (session->priv->io_channel,
-									G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL);
-    g_source_set_callback (session->priv->io_source, 
-							(GSourceFunc)obex_io_callback, 
-							obex_context->obex_handle, NULL);
-    (void) g_source_attach (session->priv->io_source, NULL);
-	
+	chan = g_io_channel_unix_new (session->priv->fd);
+	session->priv->io_watch = g_io_add_watch(chan,
+			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+			obex_io_callback, obex_context->obex_handle);
+	g_io_channel_unref(chan);
+
 	/* connect automatically */
 	if (ods_session_connect_internal (session, &error) == -1) {
 		/* error (error) */
@@ -744,12 +761,10 @@ ods_session_finalize (GObject *object)
 
 	/* close connection, free obex_context */
 	g_message ("closing connection");
-	OBEX_TransportDisconnect (session->priv->obex_context->obex_handle);
+	shutdown (session->priv->fd, SHUT_RDWR);
+	close (session->priv->fd);
 	OBEX_Cleanup (session->priv->obex_context->obex_handle);
 	g_free (session->priv->obex_context);
-	g_io_channel_shutdown (session->priv->io_channel, FALSE, NULL);
-	g_io_channel_unref(session->priv->io_channel);
-	g_source_destroy (session->priv->io_source);
 	/* free other private variables */
 	g_free (session->priv->owner);
 	g_free (session->priv->dbus_path);
@@ -844,21 +859,24 @@ ods_session_disconnect_internal (OdsSession *session, GError **error)
 	
 	if (session->priv->state == ODS_SESSION_STATE_NOT_CONNECTED) {
 		/* emit DISCONNECTED signal now */
-		g_signal_emit (session, signals[DISCONNECTED], 0);
+		g_idle_add ((GSourceFunc) emit_disconnected, session);
+		g_source_remove(session->priv->io_watch);
 		return 1;
 	}
 
 	/* actually disconnect */
+	session->priv->state = ODS_SESSION_STATE_NOT_CONNECTED;
 	ret = ods_obex_disconnect (session->priv->obex_context);
 	if (ret < 0) {
 		/* emit DISCONNECTED signal now and set state to NOT_CONNECTED
 		 * in this case disconnection will happen when socket is closed */
-		g_signal_emit (session, signals[DISCONNECTED], 0);
-		session->priv->state = ODS_SESSION_STATE_NOT_CONNECTED;
+		g_idle_add ((GSourceFunc) emit_disconnected, session);
 		
 		ods_error_err2gerror (ret, error);
+		g_source_remove(session->priv->io_watch);
 		return -1;
 	}
+
 	return 0;
 }
 
